/*
 * 
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 * 
 * Copyright 2007-2008 Sun Microsystems, Inc. All rights reserved.
 * 
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License. You can obtain
 * a copy of the License at https://glassfish.dev.java.net/public/CDDL+GPL.html
 * or glassfish/bootstrap/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 * 
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at glassfish/bootstrap/legal/LICENSE.txt.
 * Sun designates this particular file as subject to the "Classpath" exception
 * as provided by Sun in the GPL Version 2 section of the License file that
 * accompanied this code.  If applicable, add the following below the License
 * Header, with the fields enclosed by brackets [] replaced by your own
 * identifying information: "Portions Copyrighted [year]
 * [name of copyright owner]"
 * 
 * Contributor(s):
 * 
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 *
 */
/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package com.sun.grizzly.async;

import com.sun.grizzly.Controller;
import com.sun.grizzly.SelectorHandler;
import com.sun.grizzly.async.AsyncQueue.AsyncQueueEntry;
import java.io.IOException;
import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.ReadableByteChannel;
import java.nio.channels.SelectableChannel;
import java.nio.channels.SelectionKey;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicReference;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Level;

/**
 *
 * @author oleksiys
 */
public abstract class AbstractAsyncQueueReader implements AsyncQueueReader {
    private SelectorHandler selectorHandler;
    private AsyncQueue<SelectableChannel, AsyncReadQueueRecord> readQueue;
    private ConcurrentLinkedQueue<AsyncReadQueueRecord> recordQueue;
    
    public AbstractAsyncQueueReader(SelectorHandler selectorHandler) {
        this.selectorHandler = selectorHandler;
        readQueue = new AsyncQueue<SelectableChannel, AsyncReadQueueRecord>();
        recordQueue = new ConcurrentLinkedQueue<AsyncReadQueueRecord>();
    }
    
    /**
     * {@inheritDoc}
     */
    public void read(SelectionKey key, ByteBuffer buffer, 
            AsyncReadCallbackHandler callbackHandler) throws IOException {
        read(key, buffer, callbackHandler, null);
    }

    /**
     * {@inheritDoc}
     */
    public void read(SelectionKey key, ByteBuffer buffer, 
            AsyncReadCallbackHandler callbackHandler, 
            AsyncReadCondition condition) throws IOException {
        read(key, buffer, callbackHandler, condition, null);
    }
    
    /**
     * {@inheritDoc}
     */
    public void read(SelectionKey key, ByteBuffer buffer, 
            AsyncReadCallbackHandler callbackHandler, 
            AsyncReadCondition condition, 
            AsyncQueueDataProcessor readPostProcessor) throws IOException {

        if (key == null) {
            throw new IOException("SelectionKey is null! " +
                    "Probably key was cancelled or connection was closed?");
        }

        SelectableChannel channel = (SelectableChannel) key.channel();
        AsyncQueueEntry channelEntry = 
                readQueue.obtainAsyncQueueEntry(channel);
        
        ConcurrentLinkedQueue<AsyncReadQueueRecord> queue = channelEntry.queue;
        AtomicReference<AsyncReadQueueRecord> currentElement = channelEntry.currentElement;
        ReentrantLock lock = channelEntry.queuedActionLock;
        
        // If AsyncQueue is empty - try to read ByteBuffer here
        try {
            AsyncReadQueueRecord record = null;
            SocketAddress address = null;
            boolean isDirectReadCompleted = false;

            if (currentElement.get() == null && // Weak comparison for null
                    lock.tryLock()) {
                record = obtainRecord();
                // Strong comparison for null, because we're in locked region
                if (currentElement.compareAndSet(null, record)) {
                    
                    // Do direct reading
                    do {
                        address = doRead((ReadableByteChannel) channel, buffer, readPostProcessor);
                        // If some data was read - we need to check "condition"
                        // Check is performed for each message separately, not like for TCP
                        if (address != null &&
                                (!buffer.hasRemaining() || (condition != null &&
                                condition.checkAsyncReadCompleted(key, address, buffer)))) {
                            isDirectReadCompleted = true;
                            break;
                        }
                    } while (address != null);
                } else {
                    lock.unlock();
                }
            }

            if (!isDirectReadCompleted && buffer.hasRemaining()) {
                if (record == null) {
                    record = obtainRecord();
                }

                record.set(buffer, callbackHandler, condition, readPostProcessor);

                boolean isRegisterForReading = false;
                
                // add new element to the queue, if it's not current
                if (currentElement.get() != record) {
                    queue.offer(record); // add to queue
                    if (!lock.isLocked()) {
                        isRegisterForReading = true;
                    }
                } else {  // if element was read direct (not fully read)
                    isRegisterForReading = true;
                    lock.unlock();
                }
                
                if (isRegisterForReading) {
                    registerForReading(key);
                }
            } else { // If there are no bytes available for reading
                
                // Notify callback handler
                if (callbackHandler != null) {
                    callbackHandler.onReadCompleted(key,
                            address, buffer);
                }
                
                // If buffer was read directly - set next queue element as current
                if (lock.isHeldByCurrentThread()) {
                    AsyncReadQueueRecord nextRecord = queue.poll();
                    if (nextRecord != null) { // if there is something in queue
                        currentElement.set(nextRecord); 
                        lock.unlock();
                        registerForReading(key);
                    } else { // if nothing in queue
                        currentElement.set(null);
                        lock.unlock();  // unlock
                        if (queue.peek() != null) {  // check one more time
                            registerForReading(key);
                        }
                    }
                }
                
                // Release record element
                if (record != null) {
                    recordQueue.offer(record);
                }
            }
        } catch(IOException e) {
            onClose(channel);
            throw e;
        } finally {
            if (lock.isHeldByCurrentThread()) {
                lock.unlock();
            }
        }
    }
    
    /**
     * {@inheritDoc}
     */
    public boolean isAsyncQueueReaderEnabledFor(SelectionKey key) {
        AsyncQueueEntry channelEntry = 
                readQueue.getAsyncQueueEntry(key.channel());
        
        return channelEntry != null && (channelEntry.currentElement != null || 
                (channelEntry.queue != null && !channelEntry.queue.isEmpty()));
    }

    /**
     * {@inheritDoc}
     */
    public void onRead(SelectionKey key) throws IOException {
        SelectableChannel channel = key.channel();
        
        AsyncQueueEntry channelEntry = 
                readQueue.obtainAsyncQueueEntry(channel);
        
        ConcurrentLinkedQueue<AsyncReadQueueRecord> queue = channelEntry.queue;
        AtomicReference<AsyncReadQueueRecord> currentElement = channelEntry.currentElement;
        ReentrantLock lock = channelEntry.queuedActionLock;

        if (currentElement.get() == null) {
            AsyncReadQueueRecord nextRecord = queue.peek();
            if (nextRecord != null && lock.tryLock()) {
                if (!queue.isEmpty() &&
                        currentElement.compareAndSet(null, nextRecord)) {
                    queue.remove();
                }
            } else {
                return;
            }
        } else if (!lock.tryLock()) {
            return;
        }

        try {
            while (currentElement.get() != null) {
                AsyncReadQueueRecord queueRecord = currentElement.get();

                ByteBuffer byteBuffer = queueRecord.byteBuffer;
                SocketAddress address = null;
                AsyncQueueDataProcessor readPostProcessor = queueRecord.readPostProcessor;
                try {
                    address = doRead((ReadableByteChannel) channel, byteBuffer, readPostProcessor);
                } catch (IOException e) {
                    if (queueRecord.callbackHandler != null) {
                        queueRecord.callbackHandler.onIOException(e, key,
                                byteBuffer, queue);
                    } else {
                        Controller.logger().log(Level.SEVERE,
                                "Exception occured when executing " +
                                "asynchronous queue reading", e);
                    }

                    onClose(channel);
                }

                // check if buffer was completely read
                AsyncReadCondition condition = queueRecord.condition;
                if (!byteBuffer.hasRemaining() || (condition != null &&
                        condition.checkAsyncReadCompleted(key, address, byteBuffer))) {
                    if (queueRecord.callbackHandler != null) {
                        queueRecord.callbackHandler.onReadCompleted(
                                key, address, byteBuffer);
                    }

                    currentElement.set(queue.poll());
                    recordQueue.offer(queueRecord);

                    // If last element in queue is null - we have to be careful
                    if (currentElement.get() == null) {
                        lock.unlock();
                        AsyncReadQueueRecord nextRecord = queue.peek();
                        if (nextRecord != null && lock.tryLock()) {
                            if (!queue.isEmpty() &&
                                    currentElement.compareAndSet(null, nextRecord)) {
                                queue.remove();
                            }
                            
                            continue;
                        } else {
                            break;
                        }
                    }
                } else { // if there is still some data in current buffer
                    lock.unlock();
                    registerForReading(key);
                    break;
                }
            }
        } finally {
            if (lock.isHeldByCurrentThread()) {
                channelEntry.queuedActionLock.unlock();
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public void onClose(SelectableChannel channel) {
        readQueue.removeEntry(channel);
    }
    
    /**
     * {@inheritDoc}
     */
    public void close() {
        readQueue.clear();
        readQueue = null;
    }

    protected abstract SocketAddress doRead(ReadableByteChannel channel, ByteBuffer byteBuffer, 
            AsyncQueueDataProcessor readPostProcessor) throws IOException;

    private void registerForReading(SelectionKey key) {
        selectorHandler.register(key, SelectionKey.OP_READ);
    }
    
    private AsyncReadQueueRecord obtainRecord() {
        AsyncReadQueueRecord record = recordQueue.poll();
        if (record == null) {
            record = new AsyncReadQueueRecord();
        }
        
        return record;
    }
}
